from __future__ import annotations

import math
from dataclasses import dataclass
from enum import Enum
from functools import reduce
from typing import Any
from typing import Generic
from typing import List
from typing import TypeVar
from typing import Union

import numpy as np
import numpy.typing as npt
import pandas as pd

from . import plot

# A Mask is a 2D array representing the class index of each pixel of an image as values
Mask = List[List[int]]

# A Image is a 3D array representing RGB values or a 2D array representing grayscale values
Image = Union[List[List[List[int]]], List[List[int]]]

TResult = TypeVar('TResult')


def recall_string(recall: float) -> str:
    return 'No positive cases in GT' if math.isnan(recall) else str(recall)


def specificity_string(specificity: float) -> str:
    return 'No negative cases in GT' if math.isnan(specificity) else str(specificity)


@dataclass
class Result(Generic[TResult]):
    """Result of a GT versus Predictions

    Args:
        Generic (Result type): Type of the results.

    Attributes:
        image (Image): Image that the GT and predictions are based. Format: (Rows, Cols, Chanels) or (Rows, Cols) for grayscale.
        gt (TResult): Ground truth results.
        prediction (Optional[TResult]): Prediction results. None if only GT must be shown.
    """
    image: Image
    gt: TResult
    prediction: TResult | None = None


class PredictionResultType(Enum):
    TP = 0
    FP = 1
    TN = 2
    FN = 3


@dataclass
class PredictionResult:
    score: float
    type: PredictionResultType


@dataclass
class Classification:
    """Represents a classification

    Attributes:
        cls (int): Class index.
        score (Optional[float], optional): Prediction score. Defaults to None.
    """
    cls: int
    score: float | None = None


@dataclass
class BinaryClassificationMetrics:
    """Binary classification metrics

    Attributes:
        cls (str): Class name.
        TP (int): True Positive count. Defaults to 0.
        FP (int): False Positive count. Defaults to 0.
        TN (int): True Negative count. Defaults to 0.
        FN (int): False Negative count. Defaults to 0.
    """

    cls: str
    TP: int = 0
    FP: int = 0
    TN: int = 0
    FN: int = 0

    @property
    def has_instances(self) -> bool:
        """int: Indicates if the class has any ground truth or predicted instances."""
        return self.count > 0

    @property
    def count(self) -> int:
        """int: Total count of classified instances."""
        return self.TP + self.TN + self.FP + self.FN

    @property
    def accuracy(self) -> float:
        """int: Total count of classified instances."""
        if self.count == 0:
            return math.nan
        return (self.TP + self.TN) / self.count

    @property
    def recall(self) -> float:
        """float: Recall metric - TP / (TP + FN)."""
        if self.TP == 0 and self.FN == 0:
            return math.nan
        return self.TP / (self.TP + self.FN)

    @property
    def false_positive_rate(self) -> float:
        """float: False Positive Rate (FPR) metric - FP / (FP + TN)."""
        if self.FP == 0 and self.TN == 0:
            return math.nan
        return self.FP / (self.FP + self.TN)

    @property
    def specificity(self) -> float:
        """float: Specificity metric - TN / (FP + TN)."""
        if self.FP == 0 and self.TN == 0:
            return math.nan

        return self.TN / (self.FP + self.TN)

    @property
    def precision(self) -> float:
        """float: Precision metric - TP / (FP + TP)."""
        if self.FP == 0 and self.FN == 0:  # No GT instances
            return 1
        elif self.FP == 0 and self.TP == 0:
            return math.nan
        return self.TP / (self.FP + self.TP)

    @property
    def f_score(self) -> float:
        """float: F-Score/Dice metric - 2*TP / (FP + FN + 2*TP)."""
        quotient = (self.FP + self.FN + 2 * self.TP)
        if self.TP == 0 and self.FP == 0 and self.FN == 0:  # No GT instances
            return 1
        elif quotient == 0:
            return math.nan
        return 2 * self.TP / quotient

    @property
    def confusion_matrix(self) -> list[list[int]]:
        """List[List[int]]: Confusion matrix of all the classes"""
        return [[self.TP, self.FP], [self.FN, self.TN]]

    def show_confusion_matrix(self):
        """Plots de confusion matrix

        Return:
            Tuple[Figure, Axes]: Figure and Axes of the ploted confusion matrix
        """
        return plot.confusion_matrix(self.confusion_matrix, ['P', 'N'], f'Confusion Matrix for \"{self.cls}\" Class')

    def __str__(self):
        no_instances_string = ' [NO INSTANCES]' if not self.has_instances else ''
        return (
            f'{self.cls}{no_instances_string}:\n'
            f'\tTP: {self.TP}\n'
            f'\tTN: {self.TN}\n'
            f'\tFP: {self.FP}\n'
            f'\tFN: {self.FN}\n'
            f'\tFPR: {self.false_positive_rate}\n'
            f'\tAccuracy: {self.accuracy}\n'
            f'\tRecall: {recall_string(self.recall)}\n'
            f'\tPrecision: {self.precision}\n'
            f'\tSpecificity: {specificity_string(self.specificity)}\n'
            f'\tF-Score: {self.f_score}'
        )

    def to_dict(self) -> dict[str, Any]:
        if not self.has_instances:
            return {}
        else:
            return {
                'TP': self.TP,
                'TN': self.TN,
                'FP': self.FP,
                'FN': self.FN,
                'FPR': self.false_positive_rate,
                'Accuracy': self.accuracy,
                'Recall': recall_string(self.recall),
                'Precision': self.precision,
                'Specificity': specificity_string(self.specificity),
                'F-Score': self.f_score,
            }


class ClassificationMetrics:
    """Multiclass classification metrics

    Attributes:
        classes (List[str]): Class names.
        confusion_matrix (List[List[int]]): Confusion matrix of all the classes.
    """

    def __init__(self, classes: list[str], confusion_matrix: list[list[int]] | npt.NDArray[np.uint] = []):
        """
        Args:
            classes (List[str]): Class names.
            confusion_matrix (List[List[int]], optional): Confusion matrix of all the classes. Defaults to [].

        Raises:
            AssertException: If the confusion matrix is not a square matrix of order `len(classes)`
        """
        self._confusion_matrix = np.asarray(confusion_matrix, dtype=np.uint)

        assert self._confusion_matrix.shape == (
            len(classes), len(
                classes,
            ),
        ), 'The confusion matrix must be a square matrix of order len(classes)'
        self._classes = classes

        self._by_class = self.__get_by_class_metrics()
        self._by_class_w_instances = list(
            filter(
                lambda c: c.has_instances, self._by_class,
            ),
        )
        self._count = self._confusion_matrix.sum()

    @property
    def by_class(self) -> list[BinaryClassificationMetrics]:
        """List[BinaryClassificationMetrics]: Binary metrics calculated for each class index."""
        return self._by_class

    @property
    def by_class_w_instances(self) -> list[BinaryClassificationMetrics]:
        """List[BinaryClassificationMetrics]: Binary metrics calculated for each class index with instances."""
        return self._by_class_w_instances

    @property
    def count(self) -> int:
        """int: Total count of classified instances."""
        return self._count

    @property
    def accuracy(self) -> float:
        """float: Accuracy metric - correct classifications / count."""
        return np.diagonal(self._confusion_matrix).sum() / self.count

    @property
    def avg_recall(self) -> float:
        """float: Macro average recall metric."""
        by_class_w_recall = [
            c for c in self.by_class_w_instances if not math.isnan(c.recall)
        ]

        if (len(by_class_w_recall) == 0):
            return 1
        return reduce(lambda acc, curr: curr.recall + acc, by_class_w_recall, .0) / len(by_class_w_recall)

    @property
    def avg_precision(self) -> float:
        """float: Macro average precision metric."""
        return reduce(
            lambda acc, curr: (
                0 if math.isnan(curr.precision)
                else curr.precision
            ) + acc,
            self.by_class_w_instances, .0,
        ) / len(self.by_class_w_instances)

    @property
    def avg_specificity(self) -> float:
        """float: Macro average specificity metric."""
        by_class_w_specificity = [
            c for c in self.by_class_w_instances
            if not math.isnan(c.specificity)
        ]
        return reduce(lambda acc, curr: curr.specificity + acc, by_class_w_specificity, .0) / len(by_class_w_specificity)

    @property
    def avg_f_score(self) -> float:
        """float: Macro average F-Score/Dice metric."""
        return reduce(lambda acc, curr: curr.f_score + acc, self.by_class_w_instances, .0) / len(self.by_class_w_instances)

    @property
    def avg_false_positive_rate(self) -> float:
        """float: Macro average False Positive Rate metric."""
        return reduce(
            lambda acc, curr: curr.false_positive_rate + acc,
            self.by_class_w_instances, .0,
        ) / len(self.by_class_w_instances)

    @property
    def confusion_matrix(self) -> npt.NDArray[np.uint]:
        """List[List[int]]: Confusion matrix of all the classes"""
        return self._confusion_matrix

    def show_confusion_matrix(self):
        """Plots de confusion matrix

        Return:
            Tuple[Figure, Axes]: Figure and Axes of the plotted confusion
            matrix
        """
        return plot.confusion_matrix(self._confusion_matrix.tolist(), self._classes)

    def __get_by_class_metrics(self):
        by_class = []
        for i in range(0, len(self._classes)):
            tp = self._confusion_matrix[i, i]
            tn = np.delete(np.delete(self._confusion_matrix, i, 0), i, 1).sum()
            fp = self._confusion_matrix[i, :].sum() - tp
            fn = self._confusion_matrix[:, i].sum() - tp

            by_class.append(
                BinaryClassificationMetrics(
                    cls=self._classes[i],
                    TP=tp,
                    TN=tn,
                    FP=fp,
                    FN=fn,
                ),
            )
        return by_class

    def __str__(self):
        return (
            f'Classification Metrics:\n'
            f'\tSample Count: {self.count}\n'
            f'\tAccuracy: {self.accuracy}\n'
            f'\tAvg Recall: {self.avg_recall}\n'
            f'\tAvg Precision: {self.avg_precision}\n'
            f'\tAvg Specificity: {self.avg_specificity}\n'
            f'\tAvg FPR: {self.avg_false_positive_rate}\n'
            f'\tAvg F-Score: {self.avg_f_score}\n\n'
            f'By Class:\n\n'
        ) + '\n'.join([
            cls_metrics.__str__()
            for cls_metrics in self.by_class
        ])

    def to_dict(self) -> dict[str, Any]:
        return {
            'Sample Count': self.count,
            'Average': {
                'Accuracy': self.accuracy,
                'Recall': self.avg_recall,
                'Precision': self.avg_precision,
                'Specificity': self.avg_specificity,
                'FPR': self.avg_false_positive_rate,
                'F-Score': self.avg_f_score,
            },
            'By Class': dict_by_class(self.by_class),
        }

    def to_dataframe(self) -> pd.DataFrame:
        as_dict = self.to_dict()
        dict_to_df = {
            'Average': as_dict['Average'],
            **as_dict['By Class'],
        }
        return pd.DataFrame(dict_to_df)


class BinarySegmentationMetrics(BinaryClassificationMetrics):
    """Binary pixel-based classification metrics

    Attributes:
        TP (int): True Positive pixels count.
        FP (int): False Positive pixels count.
        TN (int): True Negative pixels count.
        FN (int): False Negative pixels count.
    """

    def __init__(self, classification_metrics: BinaryClassificationMetrics):
        """
        Args:
            classification_metrics (BinaryClassificationMetrics): Pixel-based binary classification metrics
        """
        super().__init__(
            cls=classification_metrics.cls,
            TP=classification_metrics.TP,
            TN=classification_metrics.TN,
            FP=classification_metrics.FP,
            FN=classification_metrics.FN,
        )

    @property
    def iou(self) -> float:
        """float: IoU/Jaccard Index metric - TP / (FP + FN + TP)."""
        quotient = (self.FP + self.FN + self.TP)
        if quotient == 0:
            return math.nan
        return self.TP / quotient

    def __str__(self):
        return (
            f'{self.cls}:\n'
            f'\tTP: {self.TP}\n'
            f'\tTN: {self.TN}\n'
            f'\tFP: {self.FP}\n'
            f'\tFN: {self.FN}\n'
            f'\tIoU: {self.iou}\n'
            f'\tAccuracy: {self.accuracy}\n'
            f'\tRecall: {recall_string(self.recall)}\n'
            f'\tPrecision: {self.precision}\n'
            f'\tSpecificity: {specificity_string(self.specificity)}\n'
            f'\tFPR: {self.false_positive_rate}\n'
            f'\tF-Score: {self.f_score}\n'
        )

    def to_dict(self) -> dict[str, Any]:
        return {
            'TP': self.TP,
            'TN': self.TN,
            'FP': self.FP,
            'FN': self.FN,
            'IoU': self.iou,
            'Accuracy': self.accuracy,
            'Recall': recall_string(self.recall),
            'Precision': self.precision,
            'Specificity': specificity_string(self.specificity),
            'FPR': self.false_positive_rate,
            'F-Score': self.f_score,
        }


class SegmentationMetrics(ClassificationMetrics):
    """Multiclass pixel-based classification metrics

    Attributes:
        classes (List[str]): Class names with `classes[0]` as the background class.
        confusion_matrix (List[List[int]]): Confusion matrix of all the classes.
    """

    def __init__(self, classes: list[str], confusion_matrix: list[list[int]] | npt.NDArray[np.uint] = []):
        """
        Args:
            classes (List[str]): Class names. It is expected the first class to be the background class.
            confusion_matrix (List[List[int]], optional): Confusion matrix of all the classes. Defaults to [].

        Raises:
            AssertException: If the confusion matrix is not a square matrix of order `len(classes)`.
            AssertException: If the number of classes is less than 2.
        """
        assert len(classes) > 1, \
            'There should be at least two classes (with background as the first)'
        super().__init__(classes, confusion_matrix)
        self._by_class = [BinarySegmentationMetrics(x) for x in self.by_class]
        self._by_class_w_instances = [
            x for x in self.by_class if x.has_instances
        ]

    @property
    def avg_iou(self) -> float:
        """float: Macro average IoU/Jaccard Index metric."""
        return reduce(lambda acc, curr: curr.iou + acc, self._by_class_w_instances, .0) / len(self._by_class_w_instances)

    @property
    def avg_iou_no_bkg(self) -> float:
        """float: Macro average IoU/Jaccard Index metric without `background` class (index 0)."""
        return reduce(
            lambda acc, curr: curr.iou + acc,
            self._by_class_w_instances[1:], .0,
        ) / (len(self._by_class_w_instances) - 1)

    def __str__(self):
        return (
            f'Segmentation Metrics:\n'
            f'\tPixel Count: {self.count}\n'
            f'\tAccuracy: {self.accuracy}\n'
            f'\tAvg Recall: {self.avg_recall}\n'
            f'\tAvg Precision: {self.avg_precision}\n'
            f'\tAvg Specificity: {self.avg_specificity}\n'
            f'\tAvg F-Score: {self.avg_f_score}\n'
            f'\tAvg FPR: {self.avg_false_positive_rate}\n'
            f'\tAvg IoU: {self.avg_iou}\n'
            f'\tAvg IoU w/o Background: {self.avg_iou_no_bkg}\n'
            f'By Class:\n\n'
        ) + '\n'.join(
            list(
                cls_metrics.__str__()
                for cls_metrics in self.by_class
            ),
        )

    def to_dict(self) -> dict[str, Any]:
        return {
            'Pixel Count': self.count,
            'Average': {
                'Accuracy': self.accuracy,
                'Recall': self.avg_recall,
                'Precision': self.avg_precision,
                'Specificity': self.avg_specificity,
                'F-Score': self.avg_f_score,
                'FPR': self.avg_false_positive_rate,
                'IoU': self.avg_iou,
                'IoU w/o Background': self.avg_iou_no_bkg,
            },
            'By Class': dict_by_class(self.by_class),
        }

    def to_dataframe(self) -> pd.DataFrame:
        as_dict = self.to_dict()
        dict_to_df = {
            'Average': as_dict['Average'],
            **as_dict['By Class'],
        }
        return pd.DataFrame(dict_to_df)


class BinaryDetectionMetrics(BinaryClassificationMetrics):
    """Binary instance-based classification metrics

    Attributes:
        TP (int): True Positive instances count.
        FP (int): False Positive instances count.
        TN (int): Not used in object detection. Always 0.
        FN (int): False Negative instances count.
    """

    def __init__(
        self,
        classification_metrics: BinaryClassificationMetrics,
        iou: float,
        predictions: list[PredictionResult],
    ):
        super().__init__(
            cls=classification_metrics.cls,
            TP=classification_metrics.TP,
            FP=classification_metrics.FP,
            FN=classification_metrics.FN,
        )
        self._iou = iou
        self._precision_recall_curve = self.__calculate_precision_recall_curve(
            predictions,
        ) if self.gt_count > 0 else []

    @property
    def gt_count(self) -> int:
        """int: Total count of GT bboxes."""
        return self.TP + self.FN

    @property
    def predicted_count(self) -> int:
        """int: Total count of predicted bboxes."""
        return self.TP + self.FP

    @property
    def iou(self) -> float:
        """float: IoU/Jaccard Index metric.

        This metric is calculated as for segmentation, considering bboxes as pixel masks.
        """
        return self._iou

    @property
    def precision_recall_curve(self) -> list[tuple[float, float]]:
        """List[Tuple[float, float]]: Precision x Recall curve as a list of (Recall, Precision) tuples."""
        assert self.gt_count > 0, 'This class does not have instances.'
        return self._precision_recall_curve

    def average_precision(self, interpolation_points: int | None = None) -> float:
        """Calculates the Average Precision metric.

        Args:
            interpolation_points (Optional[int], optional): Number of points to use for interpolation.
            Uses all points if None. Defaults to None.

        Returns:
            float: Average Precision metric.
        """
        if interpolation_points is not None:
            return self._interpolated_average_precision(interpolation_points)
        return self._average_precision()

    def show_precision_recall_curve(self):
        """Plots de Precision x Recall curve.

        Return:
            Tuple[Figure, Axes]: Figure and Axes of the ploted Precision x Recall curve.
        """
        return plot.precision_recall_curve([self._precision_recall_curve], [self.cls])

    def __calculate_precision_recall_curve(self, predictions: list[PredictionResult]) -> list[tuple[float, float]]:
        assert self.gt_count > 0, 'The GT count must be greater than 0.'

        sorted_predictions = sorted(
            predictions, key=lambda p: float(p.score), reverse=True,
        )

        curve: list[tuple[float, float]] = []
        tp = 0
        fp = 0
        gt_count = self.gt_count

        def calc_precision(tp, fp):
            return tp / (tp + fp)

        def calc_recall(tp):
            return tp / gt_count

        for prediction in sorted_predictions:
            if prediction.type == PredictionResultType.TP:
                tp += 1
            elif prediction.type == PredictionResultType.FP:
                fp += 1
            curve.append((calc_recall(tp), calc_precision(tp, fp)))

        return curve

    def _interpolated_average_precision(self, interpolation_points: int) -> float:
        precision_sum = .0
        recall_interpolation = np.linspace(0, 1, interpolation_points)

        for point in recall_interpolation:
            precision_sum += max([
                rp[1]
                for rp in self._precision_recall_curve if rp[0] >= point
            ] + [0])

        return precision_sum / interpolation_points

    def _average_precision(self) -> float:
        ap_sum = .0
        prev_max_recall = .0

        greater_recall_points = self._precision_recall_curve
        while len(greater_recall_points) > 0:
            max_point = max(greater_recall_points, key=lambda rp: rp[1])
            ap_sum += (max_point[0] - prev_max_recall) * max_point[1]
            prev_max_recall = max_point[0]
            greater_recall_points = [
                rp for rp in greater_recall_points if rp[0] > prev_max_recall
            ]

        return ap_sum

    def __str__(self):
        return (
            f'{self.cls}:\n'
            f'\tTP: {self.TP}\n'
            f'\tFP: {self.FP}\n'
            f'\tFN: {self.FN}\n'
            f'\tTN: [NA]\n'
            f'\tIoU: {self.iou}\n'
            f'\tAccuracy: {self.accuracy}\n'
            f'\tRecall: {recall_string(self.recall)}\n'
            f'\tPrecision: {self.precision}\n'
            f'\tF-Score: {self.f_score}\n'
            f'\tAverage Precision: {self.average_precision()}\n'
            f'\t11-point Average Precision: {self.average_precision(11)}\n'
        )

    def to_dict(self) -> dict[str, Any]:
        return {
            'TP': self.TP,
            'FP': self.FP,
            'FN': self.FN,
            'TN': math.nan,
            'IoU': self.iou,
            'Accuracy': self.accuracy,
            'Recall': recall_string(self.recall),
            'Precision': self.precision,
            'F-Score': self.f_score,
            'Average Precision': self.average_precision(),
            '11-point Average Precision': self.average_precision(11),
        }


class DetectionMetrics(ClassificationMetrics):
    """Multiclass detection metrics

    Attributes:
        classes (List[str]): Class names. The last must be the "undetected" class.
        confusion_matrix (List[List[int]]): Confusion matrix of all the classes.
        iou_by_class (List[float]): IoU values indexed by class.

        The last column and line must correspond to the "undetected" class.
    """

    def __init__(
        self,
        classes: list[str],
        confusion_matrix: list[list[int]] | npt.NDArray[np.uint],
        iou_by_class: list[float],
        predictions_by_class: list[list[PredictionResult]],
    ):
        super().__init__(classes, confusion_matrix)
        by_class_wo_undetected = self.by_class[slice(0, -1)]
        self._by_class = [
            BinaryDetectionMetrics(by_class, iou, predictions)
            for by_class, iou, predictions in zip(by_class_wo_undetected, iou_by_class, predictions_by_class)
        ]
        self._by_class_w_instances = [
            x for x in self.by_class if x.has_instances
        ]

    @property
    def avg_iou(self):
        """float: Macro average IoU/Jaccard Index metric.

        This metric is calculated as for segmentation, considering bboxes as pixel masks.
        """
        return reduce(lambda acc, curr: curr.iou + acc, self._by_class_w_instances, .0) / len(self._by_class_w_instances)

    def mean_average_precision(self, interpolation_points: int | None = None) -> float:
        """Calculates the mean of Average Precision metrics of all classes.

        Args:
            interpolation_points (Optional[int], optional): Number of points to use for interpolation.
            Uses all points if None. Defaults to None.

        Returns:
            float: Mean Average Precision metric.
        """
        return reduce(
            lambda acc,
            curr: curr.average_precision(interpolation_points) + acc, self._by_class_w_instances,
            .0,
        ) / len(self._by_class_w_instances)

    def __str__(self):
        return (
            f'Detection Metrics:\n'
            f'\tBboxes Count: {self.count}\n'
            f'\tAccuracy: {self.accuracy}\n'
            f'\tAvg Recall: {self.avg_recall}\n'
            f'\tAvg Precision: {self.avg_precision}\n'
            f'\tAvg F-Score: {self.avg_f_score}\n'
            f'\tAvg IoU: {self.avg_iou}\n'
            f'By Class:\n\n'
        ) + '\n'.join(
            list(
                cls_metrics.__str__()
                for cls_metrics in self.by_class
            ),
        )

    def to_dict(self) -> dict[str, Any]:
        return {
            'Bboxes Count': self.count,
            'Average': {
                'Accuracy': self.accuracy,
                'Recall': self.avg_recall,
                'Precision': self.avg_precision,
                'F-Score': self.avg_f_score,
                'IoU': self.avg_iou,
            },
            'By Class': dict_by_class(self.by_class),
        }

    def to_dataframe(self) -> pd.DataFrame:
        as_dict = self.to_dict()
        dict_to_df = {
            'Average': as_dict['Average'],
            **as_dict['By Class'],
        }
        return pd.DataFrame(dict_to_df)


def dict_by_class(by_class) -> dict[str, Any]:
    return {
        cls_metrics.cls: cls_metrics.to_dict() for cls_metrics in by_class
    }



from __future__ import annotations

import argparse
import os
from typing import Sequence

from typing_extensions import TypeAlias

from lapixdl.convert import save_lapixdf_as_masks
from lapixdl.formats import lapix


SUBPARSER_T: TypeAlias = 'argparse._SubParsersAction[argparse.ArgumentParser]'


def converter_command_parser(subparsers: SUBPARSER_T | None = None) -> argparse.ArgumentParser:
    if subparsers is not None:
        parser = subparsers.add_parser('converter')
    else:
        parser = argparse.ArgumentParser(prog='lapixdl dataset converter command')

    group_output = parser.add_argument_group('Define the output format desired')
    mexg_output = group_output.add_mutually_exclusive_group(required=True)
    mexg_output.add_argument(
        '--to-masks',
        action='store_true',
        help='Convert the input data into semantic segmentation masks. This conversion is allowed from lapixdl format.',
    )

    converter_group = parser.add_argument_group('General parameters for the converters command')
    converter_group.add_argument('-i', '--in-path', help='Path for the input file.', required=True, metavar='INPUT_PATH')
    converter_group.add_argument('-o', '--out-path', help='Path for the output dir.', required=True, metavar='OUTPUT_DIR')
    converter_group.add_argument('-j', '--jobs', help='Number of processes/jobs to be used.', default=1)

    masks_group = parser.add_argument_group('Parameters for the conversion of masks')
    masks_group.add_argument('--draw-order', nargs='+', type=int, default=None)
    masks_group.add_argument('--mask-extension', type=str, default='.png')

    if subparsers is not None:
        parser.set_defaults(func=converter_command)

    return parser


def to_masks(
    in_path: str,
    out_path: str,
    mask_extension: str,
    draw_order: tuple[int, ...] | None,
    processes: int,
) -> int:

    print(f'Loading annotations from {in_path}')
    lapix_df = lapix.load(in_path)

    print('Generating the masks...')
    save_lapixdf_as_masks(lapix_df, out_path, mask_extension, draw_order, processes)

    return 0


def converter_command(args: argparse.Namespace | None = None) -> int:
    if args is None:
        return 1

    # if args.to_masks:
    draw_order = tuple(args.draw_order) if isinstance(args.draw_order, list) else None

    return to_masks(
        in_path=os.path.abspath(args.in_path),
        out_path=os.path.abspath(args.out_path),
        mask_extension=args.mask_extension,
        draw_order=draw_order,
        processes=args.jobs,
    )


def main(argv: Sequence[str] | None = None) -> int:
    parser = converter_command_parser()
    args = parser.parse_args(argv)
    return converter_command(args)


if __name__ == '__main__':
    raise SystemExit(main())



from __future__ import annotations

import os


class FileTypeError(RuntimeError):
    """A runtime error when an unexpected file type was encountered."""
    pass


def basename(
    filename: str,
    with_extension: bool = False,
) -> str:
    """Gets the basename from a full filename

    Args:
        filename (str): The full filename of the file
        with_extension (bool): To determine if the return is the basename
    with (True) or without the extension (False).

    Returns:
        str: The basename from the filename.
    """

    bn = os.path.basename(filename)
    if with_extension:
        return bn
    else:
        return os.path.splitext(bn)[0]



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import numpy as np
import numpy.typing as npt
from PIL import Image


@dataclass
class Mask:
    # FIXME: typing also can be a list[list[int]]
    categorical: npt.NDArray[np.uint8]

    def __post_init__(self) -> None:
        self.categorical = np.array(self.categorical, dtype=np.uint8)

        if not len(self.categorical.shape) == 2:
            raise ValueError('Unexpected shape. The categorical mask needs to be a 2D array.')

    @property
    def height(self) -> int:
        return int(self.categorical.shape[0])

    @property
    def width(self) -> int:
        return int(self.categorical.shape[1])

    @property
    def unique_ids(self) -> set[int]:
        return set(np.unique(self.categorical))

    def save(
        self,
        filename: str,
        **kwargs: Any,
    ) -> None:
        Image.fromarray(self.categorical).save(filename, **kwargs)



from __future__ import annotations

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sn
from matplotlib.axes import Axes
from matplotlib.figure import Figure


def confusion_matrix(
    confusion_matrix: list[list[int]],
    classes: list[str],
    title: str | None = 'Confusion Matrix',
) -> tuple[Figure, Axes]:
    """Plots a confusion matrix

    Args:
        confusion_matrix (List[List[int]]): Confusion matrix.
        classes (List[str]): Class names.
        title (Optional[str], optional): Chart title. Defaults to
        "Confusion Matrix".

    Return:
        Tuple[Figure, Axes]: Figure and Axes of the plotted confusion
        matrix.
    """

    df_cm = pd.DataFrame(confusion_matrix, classes, classes)

    fig, ax = plt.subplots()
    fig.suptitle(title)

    ax = sn.heatmap(
        df_cm,
        annot=True,
        ax=ax,
        cmap='YlGnBu',
    )

    ax.set_xlabel('Actual')
    ax.set_ylabel('Predicted')
    plt.show()

    return fig, ax


def precision_recall_curve(
    recalls_precisions_by_class: list[list[tuple[float, float]]],
    classes: list[str],
    title: str | None = 'Precision x Recall Curve',
) -> tuple[Figure, Axes]:
    """Plots a Precision x Recall curve.

    Args:
        recalls_precisions_by_class (List[List[Tuple[float, float]]]): (Recall, Precision) tuples indexed by class.
        classes (List[str]): Class names.
        title (Optional[str], optional): Chart title. Defaults to "Precision x Recall Curve".

    Return:
        Tuple[Figure, Axes]: Figure and Axes of the ploted Precision x Recall curve.
    """

    table = []
    for i, recalls_precisions_cls in enumerate(recalls_precisions_by_class):
        for recall_precision in recalls_precisions_cls:
            table.append(
                (classes[i], recall_precision[0], recall_precision[1]),
            )

    df = pd.DataFrame(table, columns=['class', 'recall', 'precision'])

    fig, ax = plt.subplots()
    fig.suptitle(title)

    ax = sn.lineplot(
        data=df,
        y='precision',
        x='recall',
        hue='class',
        ax=ax,
    )

    ax.set_xlabel('Recall')
    ax.set_ylabel('Precision')
    plt.show()

    return fig, ax



from __future__ import annotations

import multiprocessing
import os
import sys
from collections import defaultdict
from itertools import chain
from typing import Any
from typing import Iterator

import numpy as np
import pandas as pd
from PIL import Image
from PIL import ImageDraw
from shapely.geometry import MultiPolygon
from shapely.geometry import Point
from shapely.geometry import Polygon

from lapixdl.base import basename
from lapixdl.formats.annotation import Annotation
from lapixdl.formats.lapix import LapixDataFrame
from lapixdl.formats.mask import Mask


# -----------------------------------------------------------------------
# Functions to work with data from LabelBox (lbox)
def lbox_geo_to_shapely(object: dict[str, Any]) -> Polygon | Point | float:
    """Convert the labelbox geometries into shapely geometries
    (For polygons or Points)
    """
    keys = object.keys()

    if 'polygon' in keys:
        polygon = object['polygon']
        geometry = Polygon(np.array([(p['x'], p['y']) for p in polygon]))
    elif 'point' in keys:
        point = object['point']
        geometry = Point(np.array([point['x'], point['y']]))
    else:
        geometry = np.NaN
    return geometry


def __lbox_has_review(reviews: list[dict[str, Any]]) -> bool:
    """Verify if the labelbox review field has at least one review"""
    for x in reviews:
        if x['score'] == 1:
            return True
    else:
        return False


def __lbox_drop_duplicate_labels(labelbox_df: pd.DataFrame) -> pd.DataFrame:
    """Remove duplicated annotations from labelbox dataframe"""

    duplicated_idx = labelbox_df['image_name'].duplicated(keep=False)
    df_duplicated = labelbox_df.loc[duplicated_idx, :].copy()

    if df_duplicated.empty:
        return labelbox_df

    # Check the labels that has review
    df_duplicated['have_review'] = df_duplicated.apply(lambda row: __lbox_has_review(row['Reviews']), axis=1)

    # Count the quantity of labels for each row
    df_duplicated['len'] = df_duplicated.apply(lambda row: len(row['Label']['objects']), axis=1)

    # Sort the DF by the quantity of labels
    df_duplicated = df_duplicated.sort_values(['image_name', 'len'], ascending=False)

    # Drop the duplicates labels and keep the first label will be that have more labels
    df_to_keep = df_duplicated.drop_duplicates(['image_name'], keep='first')

    id_to_remove = df_duplicated.loc[~df_duplicated['ID'].isin(df_to_keep['ID'].to_numpy()), 'ID']
    # the rows without review
    id_to_remove = pd.concat([id_to_remove, df_duplicated.loc[~df_duplicated['have_review'], 'ID']])

    df_without_duplicated = labelbox_df[~labelbox_df['ID'].isin(id_to_remove)].copy()

    return df_without_duplicated


def __lbox_explode_images_annotations(labelbox_df: pd.DataFrame) -> pd.DataFrame:
    """Explode image annotations of the Labelbox dataframe"""
    labelbox_df['objects'] = labelbox_df.apply(lambda row: row['Label']['objects'], axis=1)
    labelbox_df = labelbox_df.explode('objects')
    labelbox_df = labelbox_df.reset_index()

    labelbox_df = labelbox_df.drop(['index', 'Label'], axis=1)
    return labelbox_df


def __lbox_cast_geometries(labelbox_df: pd.DataFrame) -> pd.DataFrame:
    """Cast the labelbox geometries into shapely geometries"""
    labelbox_df['geometry'] = labelbox_df['objects'].apply(lambda obj: lbox_geo_to_shapely(obj))
    df_out = labelbox_df.dropna(axis=0, subset=['geometry'])

    if labelbox_df.shape != df_out.shape:
        print(f'Some NaN geometries have been deleted! Original shape = {labelbox_df.shape} | out shape = {df_out.shape}')

    if df_out.empty:
        raise ValueError('Data without valid geometries! After transform the geometries the dataframe stay empty.')

    return df_out


def labelbox_to_lapix(
    labelbox_df: pd.DataFrame,
    schematic_to_id: dict[str, int],
) -> LapixDataFrame:
    '''Transform the raw dataframe from LabelBox data to Lapix dataframe'''

    df_out = labelbox_df.copy()

    # Drop ignored images at labelling process
    if 'Skipped' in df_out.columns:
        df_out = df_out.drop(df_out[df_out['Skipped']].index)

    # Drop irrelevant columns
    df_out = df_out.drop(
        [
            'DataRow ID', 'Labeled Data', 'Created By', 'Project Name', 'Dataset Name', 'Created At', 'Updated At',
            'Seconds to Label', 'Agreement', 'Benchmark Agreement', 'Benchmark ID', 'View Label',
            'Has Open Issues', 'Skipped',
        ], axis=1, errors='ignore',
    )

    # Get image names
    df_out['image_name'] = df_out.apply(lambda row: basename(row['External ID']), axis=1)
    df_out = df_out.drop(['External ID'], axis=1)

    # Remove duplicated labels
    df_out = __lbox_drop_duplicate_labels(df_out)

    # Explode annotations to each row
    df_out = __lbox_explode_images_annotations(df_out)

    # Transform labelbox annotation to a geometry
    df_out = __lbox_cast_geometries(df_out)

    # Map category IDs
    df_out['category_id'] = df_out.apply(lambda row: schematic_to_id[row['objects']['schemaId']], axis=1)

    df_out = df_out.drop(['ID', 'objects', 'Reviews'], axis=1)

    lapix_df = LapixDataFrame(df_out)

    return lapix_df


# -----------------------------------------------------------------------
# Functions work with data to COCO
def bounds_to_coco_bb(
    bounds: tuple[float],
    decimals: int = 2,
) -> list[float]:
    """Generate a bbox in coco format from a tuple of bounds

    Args:
        bounds (tuple[float]): A tuple of floats in the following order
    (min_x, min_y, max_x, max_y)
        decimals: The quantity of decimals to round the data

    Returns:
        list[float]: An bbox into coco format based on the bounds
    """
    # bounds is in  (minx, miny, maxx, maxy)
    # bb  of coco is in [min(x), min(y), max(x)-min(x), max(y)-min(y)]
    b = tuple(np.round(x, decimals) for x in bounds)
    min_x, min_y, max_x, max_y = b
    return [min_x, min_y, max_x - min_x, max_y - min_y]


def pol_to_coco_segment(
    geo: Polygon | MultiPolygon,
    decimals: int = 2,
) -> list[list[float]]:
    """Generate a coco segment from shapely polygon"""
    # polygon of shapely is a class
    # polygon or segmentation at coco is a list of [[x0, y0, x1, y1 ...]]

    def coco_pol(geometry: Polygon, decimals: int = decimals) -> list[float]:
        list_of_points = list(zip(*geometry.exterior.coords.xy))
        geometry = list(chain(*list_of_points))
        geometry = list(np.around(np.array(geometry), decimals))
        return geometry

    if geo.geom_type == 'Polygon':
        return [coco_pol(geo)]
    elif geo.geom_type == 'MultiPolygon':
        return [coco_pol(g) for g in geo.geoms]
    else:
        raise TypeError(f'Geometry shape is not a polygon or MultiPolygon. This is a {geo.geom_type}.')


def __lapix_to_od_coco_annotations(
    lapix_df: LapixDataFrame,
    decimals: int = 2,
) -> list[dict[str, Any]]:
    """Convert lapix dataframe annotations into object detection
    annotations in a single process"""
    cols = lapix_df.columns
    if not all(c in cols for c in ['area', 'image_id', 'iscrowd']):
        raise KeyError('The dataframe need to have the columns `area`, `image_id`, `iscrowd`!')

    return lapix_df.apply(
        lambda row: {
            'id': row.name,
            'image_id': row['image_id'],
            'category_id': row['category_id'],
            'bbox': bounds_to_coco_bb(row['geometry'].bounds, decimals),
            'segmentation': pol_to_coco_segment(row['geometry'], decimals),
            'area': np.round(row['area'], decimals),
            'iscrowd': row['iscrowd'],
        },
        axis=1,
    ).to_numpy().tolist()


def lapix_to_od_coco_annotations(
    lapix_df: LapixDataFrame,
    decimals: int = 2,
    processes: int = 1,
) -> list[dict[str, Any]]:
    """Convert lapix dataframe annotations into object detection
    annotations in a multiprocesses

    Args:
        lapix_df (LapixDataFrame): The data into Lapix format, need to
    have the columns: `area`, `image_id`, `iscrowd` and `geometry`
        decimals (int): The quantity of decimals desired on the coco
    annotations
        processes (int): The number of processes to be triggered to
    perform the conversion

    Results:
        list[dict[str, Any]]: A list with the annotations in the coco
    for object detection format.
    """

    cols = lapix_df.columns
    if not all(c in cols for c in {'area', 'image_id', 'iscrowd', 'geometry'}):
        raise KeyError('The dataframe need to have the columns `area`, `image_id`, `iscrowd` and `geometry`!')

    # To ensure that will have sequential index
    lapix_df.reset_index(drop=True, inplace=True)
    lapix_df.index = lapix_df.index + 1

    # Split equally the annotations for the processes quantity
    ann_ids_splitted = np.array_split(lapix_df.index.tolist(), processes)
    print(f'Number of processes: {processes}, annotations per process: {len(ann_ids_splitted[0])}')

    workers = multiprocessing.Pool(processes=processes)
    procs = []
    for ann_ids in ann_ids_splitted:
        df_to_process = lapix_df.loc[lapix_df.index.isin(ann_ids), :]
        p = workers.apply_async(__lapix_to_od_coco_annotations, (df_to_process, decimals))
        procs.append(p)

    annotations_coco = []
    for p in procs:
        annotations_coco.extend(p.get())

    return annotations_coco


def __generate_coco_od_file(
    lapix_df: LapixDataFrame,
    categories_coco: list[dict[str, str]],
    decimals: int = 2,
    processes: int = 1,
    info_coco: dict[str, Any] | None = None,
) -> dict[str, Any]:
    """Create the object detection COCO data

    Args:
        lapix_df (LapixDataFrame): The data into Lapix format, need to
    have the columns: `area`, `image_id`, `iscrowd` and `geometry`.
        categories_coco (list[dict[str, str]]): A list with dicts within
    the categories into coco format.
        decimals (int): The quantity of decimals desired on the coco
    annotations
        processes (int): The number of processes to be triggered to
    perform the conversion
        info_coco (dict[str, Any] | None ): The data to be the `info`
    field of the COCO data.
    """
    lapix_df_clean = lapix_df.drop_duplicates('image_name')

    images_coco = [
        {
            'file_name': row['image_name'],
            'height': row['image_height'],
            'width': row['image_width'],
            'id': row['image_id'],
        } for _, row in lapix_df_clean.iterrows()
    ]

    annotations_coco = lapix_to_od_coco_annotations(lapix_df, decimals, processes)

    coco_od = {
        'categories': categories_coco,
        'images': images_coco,
        'annotations': annotations_coco,
        'info': info_coco if isinstance(info_coco, dict) else {},
    }

    return coco_od


# -----------------------------------------------------------------------
# Functions to work with lapix DataFrame
def lapix_to_annotations(lapix_df: LapixDataFrame) -> list[Annotation]:
    return [
        Annotation(
            row['geometry'],
            row['category_id'],
            row['iscrowd'] if 'iscrowd' in row else 0,
        )
        for _, row in lapix_df.iterrows()
    ]


def lapix_to_masks(
    lapix_df: LapixDataFrame,
    draw_order: tuple[int, ...] | None = None,
) -> Iterator[tuple[str, Mask]]:
    df_groupped = lapix_df.groupby('image_id')
    for _, df_by_img in df_groupped:
        annotations = lapix_to_annotations(df_by_img)

        w = df_by_img['image_width'].unique()[0]
        h = df_by_img['image_height'].unique()[0]
        image_name = basename(df_by_img.iloc[0]['image_name'])

        yield (
            image_name,
            annotations_to_mask(annotations, int(w), int(h), draw_order),
        )


# -----------------------------------------------------------------------
# Functions to generate masks
def sort_annotations_to_draw(
    annotations: list[Annotation],
    draw_order: tuple[int, ...] | None = None,
) -> list[Annotation]:
    '''Sorts the annotations based on the draw_order. If draw_order
    is None sorts in ascending order based on the category value'''
    items = defaultdict(list)

    for ann in annotations:
        items[ann.category_id].append(ann)

    if draw_order is None:
        draw_order = tuple(np.unique(list(items)))
    elif not isinstance(draw_order, tuple):
        raise ValueError('Unexpected value for `draw_order`. The value needs to be a tuple of int or None.')

    out = [ann for cat in draw_order for ann in items[cat]]

    return out


def draw_annotation(
    target: Image.Image,
    annotation: Annotation,
    value: int | tuple[int, int, int],
) -> Image.Image:
    '''Image: Draw an Annotation into the target (an Image)'''
    for geo in annotation:
        pol_x, pol_y = geo.exterior.coords.xy
        coords = list(zip(pol_x, pol_y))
        ImageDraw.Draw(target).polygon(coords, fill=value)

    return target


def annotations_to_mask(
    annotations: list[Annotation],
    width: int = 1600,
    height: int = 1200,
    draw_order: tuple[int, ...] | None = None,
) -> Mask:
    """Creates a Mask with the specified shape based on the annotations"""
    shape = (height, width)

    annotations_sorted = sort_annotations_to_draw(annotations, draw_order)

    out = Image.fromarray(np.zeros(shape, dtype=np.uint8))
    for ann in annotations_sorted:
        out = draw_annotation(out, ann, ann.category_id)

    return Mask(np.array(out))


def __save_masks_as_files(
    lapix_df: LapixDataFrame,
    output_directory: str,
    mask_extension: str = '.png',
    draw_order: tuple[int, ...] | None = None,
) -> None:

    for image_name, mask in lapix_to_masks(lapix_df, draw_order):
        out_path = os.path.join(
            output_directory,
            image_name + mask_extension,
        )
        mask.save(out_path)


def save_lapixdf_as_masks(
    lapix_df: LapixDataFrame,
    output_directory: str,
    mask_extension: str = '.png',
    draw_order: tuple[int, ...] | None = None,
    processes: int = 1,
) -> None:
    '''Saves masks as files based on the annotations from a Lapix DataFrame

    Args:
        lapix_df (LapixDataFrame): The data in the Lapix format. Needs to
    have the following columns: `image_id` and `geometry`.
        output_directory (str): The directory where the images should be
    saved.
        mask_extension (str): The image file extension/type. Defaults to
    '.png'.
        draw_order (tuple[int, ...] | None): If draw_order is None sorts
    in ascending order based on the category value. Otherwise this will
    draw the masks based on the tuple sequence.
        processes (int): The number of processes to be used to perform
    the conversion.

    '''
    img_ids = lapix_df['image_id'].unique()

    if len(img_ids) == 0:
        print('There is no annotation to save as mask.', file=sys.stderr)
        return

    images_ids_splitted = np.array_split(img_ids, processes)
    print(
        f'Starting to save {len(img_ids)} semantic segmentation masks using '
        f'{processes} processes with {len(images_ids_splitted[0])} masks per process...',
    )

    workers = multiprocessing.Pool(processes=processes)
    procs = []

    for images_ids in images_ids_splitted:
        df_to_process = lapix_df.loc[lapix_df['image_id'].isin(images_ids), :]
        p = workers.apply_async(__save_masks_as_files, (df_to_process, output_directory, mask_extension, draw_order))
        procs.append(p)

    workers.close()
    workers.join()


# -----------------------------------------------------------------------
# Functions to convert from x to y directly (ex: labelbox to coco)
def labelbox_to_coco(
    labelbox_filename: str,
    schematic_to_id: dict[str, int],
    categories_coco: list[dict[str, str]],
    *,
    target: str = 'OD',
    decimals: int = 2,
    processes: int = 1,
    info_coco: dict[str, Any] | None = None,
    image_shape: tuple[int, int] | None = None,
) -> dict[str, Any]:
    """Generate dictionary into COCO format from labelbox

    Args:
        labelbox_filename (str): The full filename for the labelbox
    JSON file
        schematic_to_id (dict[str, int]): A map dictionary between
    schematic id at labelbox and the matching category
        categories_coco (list[dict[str, str]]): A list with dicts within
    the categories into coco format.
        target (str): The desired target format of the COCO file, this
    can be (`OD` or `OBJECT DETECTION`)
        decimals (int): The quantity of decimals desired on the coco
    annotations
        processes (int): The number of processes to be triggered to
    perform the conversion
        info_coco (dict[str, Any] | None ): The data to be the `info`
    field of the COCO data.
        image_shape (tuple[int, int] | None): If a tuple will set this
    shape (height, width) for all images, if None, will try open and
    load the shape of each image.

    Return:
        dict[str, Any]: A dictionary with the data into the COCO target
    format
    """
    from lapixdl.formats import labelbox
    from lapixdl.formats import lapix

    labelbox_df = labelbox.load(labelbox_filename)

    lapix_df = labelbox_to_lapix(labelbox_df, schematic_to_id)

    lapix_df['image_id'] = lapix.generate_ids(lapix_df['image_name'])
    lapix_df['area'] = lapix.geometries_area(lapix_df)
    lapix_df['iscrowd'] = 0  # TODO: Auto check this?

    if isinstance(image_shape, tuple):
        height, width = image_shape[:2]
        lapix_df['image_height'] = height
        lapix_df['image_width'] = width
    else:
        raise NotImplementedError

    if target.upper() in {'OD', 'OBJECT DETECTION'}:
        coco_dict = __generate_coco_od_file(
            lapix_df,
            categories_coco,
            decimals,
            processes,
            info_coco,
        )
    else:
        raise NotImplementedError

    return coco_dict



from __future__ import annotations

import itertools
import warnings
from typing import Iterable

import numpy as np
import numpy.typing as npt
from tqdm.auto import tqdm

from .model import Classification
from .model import ClassificationMetrics
from .model import DetectionMetrics
from .model import Mask
from .model import PredictionResult
from .model import PredictionResultType
from .model import SegmentationMetrics
from lapixdl.formats.annotation import BBox


def evaluate_segmentation(
    gt_masks: Iterable[Mask],
    pred_masks: Iterable[Mask],
    classes: list[str],
) -> SegmentationMetrics:
    """Evaluates segmentation predictions

    The iterables should return one mask per iteration and the itens of
    `gt_masks` and `pred_masks` with same index should correspond to the
    same sample.

    Masks should be 2D arrays where each value corresponds to the class
    index of the pixel the sample image.

    Args:
        gt_masks (Iterable[Mask]): Iterable of ground truth masks.
        pred_masks (Iterable[Mask]): Iterable of predicted masks.
        classes (List[str]): Class names.

    Returns:
        SegmentationMetrics: Pixel-based classification and segmentation metrics.
    """

    classes_count = len(classes)
    zeros_matrix = np.zeros((classes_count, classes_count), dtype=np.uint)
    confusion_matrix = zeros_matrix.copy()
    classes_count_range = range(classes_count)

    for (curr_gt_mask, curr_pred_mask) in tqdm(zip(gt_masks, pred_masks), unit=' masks'):
        curr_gt_mask = np.array(curr_gt_mask)
        curr_pred_mask = np.array(curr_pred_mask)
        if curr_gt_mask.shape != curr_pred_mask.shape:
            warnings.warn(
                f'The GT mask and Pred mask should have the same shape. GT shape: {curr_gt_mask.shape}.'
                f'Pred shape: {curr_pred_mask.shape}.',
            )

        curr_confusion_matrix = zeros_matrix.copy()
        for i, j in itertools.product(classes_count_range, classes_count_range):
            curr_confusion_matrix[j, i] = np.sum((curr_pred_mask == j) * (curr_gt_mask == i))

        confusion_matrix += curr_confusion_matrix

    return SegmentationMetrics(classes, confusion_matrix)


def evaluate_classification(
    gt_classifications: Iterable[Classification],
    pred_classifications: Iterable[Classification],
    classes: list[str],
) -> ClassificationMetrics:
    """Evaluates classification predictions

    The iterables should return one classification per iteration and
    the itens of `gt_classifications` and `pred_classifications` with
    same index should correspond to the same sample.

    Args:
        gt_classifications (Iterable[Classification]): Ground truth classifications.
        pred_classifications (Iterable[Classification]): Predicted classifications.
        classes (List[str]): Class names.

    Returns:
        ClassificationMetrics: Classification metrics.
    """

    confusion_matrix = np.zeros((len(classes), len(classes)), int)

    for (curr_gt_classification, curr_pred_classification) in tqdm(
        zip(
            gt_classifications,
            pred_classifications,
        ),
        unit=' samples',
    ):
        confusion_matrix[
            curr_pred_classification.cls,
            curr_gt_classification.cls,
        ] += 1

    metrics = ClassificationMetrics(classes, confusion_matrix)

    return metrics


def evaluate_detection(
    gt_bboxes: Iterable[list[BBox]],
    pred_bboxes: Iterable[list[BBox]],
    classes: list[str],
    iou_threshold: float = .5,
    undetected_cls_name: str = '_undetected_',
) -> DetectionMetrics:
    """Evaluates detection predictions

    The iterables should return the bboxes of one image per iteration
    and the itens of `gt_bboxes` and `pred_bboxes` with same index
    should correspond to the same image.

    Args:
        gt_bboxes (Iterable[List[BBox]]): Ground truth classifications.
        pred_bboxes (Iterable[List[BBox]]): Predicted classifications.
        classes (List[str]): Class names.
        iou_threshold (float): Minimum IoU threshold to consider a detection as a True Positive.
        undetected_cls_name (str): Name to be used for undetected instances class. Defaults to "_undetected_".

    Returns:
        DetectionMetrics: Detection metrics.
    """
    tot_images = 0
    classes_count = len(classes)

    cls_ious_sum = np.zeros(classes_count)
    confusion_matrix = np.zeros((classes_count + 1, classes_count + 1), dtype=np.uint)
    undetected_idx = classes_count
    # Tracks detection scores to calculate the Precision x Recall curve and the Average Precision metric
    predictions_by_class: list[list[PredictionResult]] = [
        [] for i in range(len(classes))
    ]

    # For each image GT and predicted bbox set
    for (curr_gt_bboxes, curr_pred_bboxes) in tqdm(zip(gt_bboxes, pred_bboxes), unit=' samples'):
        pairwise_bbox_ious = calculate_pairwise_bbox_ious(
            curr_gt_bboxes, curr_pred_bboxes,
        )

        # Sums classes' IoUs by image to calculate the average at the end
        curr_img_cls_ious = calculate_iou_by_class(
            curr_gt_bboxes, curr_pred_bboxes, classes_count,
        )
        cls_ious_sum = np.add(cls_ious_sum, curr_img_cls_ious)
        tot_images += 1

        # Tracks prediction bboxes that does not correspond to any GT bbox to identify FPs
        no_hit_idxs = set(range(0, len(curr_pred_bboxes)))

        # Iterate through lines of predictions bbox IoUs for each GT bbox
        for i, gt_ious in enumerate(pairwise_bbox_ious):
            gt_cls_idx = curr_gt_bboxes[i].cls

            # Removes already matched predictions
            gt_ious = [
                0 if i not in no_hit_idxs else iou for i,
                iou in enumerate(gt_ious)
            ]

            if len(gt_ious) > 0:
                max_iou_idx = np.argmax(gt_ious)
                # Only the max IoU is considered TP, the others are FPs
                max_iou = gt_ious[max_iou_idx]
            else:
                max_iou = 0

            if max_iou < iou_threshold:  # FN - GT bbox not detected
                confusion_matrix[undetected_idx, gt_cls_idx] += 1
            else:  # TP - GT bbox detected
                no_hit_idxs.remove(int(max_iou_idx))  # Remove from FPs

                pred_cls_idx = curr_pred_bboxes[max_iou_idx].cls
                confusion_matrix[pred_cls_idx, gt_cls_idx] += 1
                if pred_cls_idx == gt_cls_idx:
                    predictions_by_class[pred_cls_idx]\
                        .append(PredictionResult(curr_pred_bboxes[max_iou_idx].score, PredictionResultType.TP))
                else:
                    predictions_by_class[pred_cls_idx]\
                        .append(PredictionResult(curr_pred_bboxes[max_iou_idx].score, PredictionResultType.FP))

        for no_hit_idx in no_hit_idxs:  # FPs - Predictions that do not match any GT
            pred_cls_idx = curr_pred_bboxes[no_hit_idx].cls
            confusion_matrix[pred_cls_idx, undetected_idx] += 1
            predictions_by_class[pred_cls_idx]\
                .append(PredictionResult(curr_pred_bboxes[no_hit_idx].score, PredictionResultType.FP))

    metrics = DetectionMetrics(
        classes + [undetected_cls_name],
        confusion_matrix,
        (cls_ious_sum / tot_images).tolist(),
        predictions_by_class,
    )

    return metrics


def calculate_pairwise_bbox_ious(
    gt_bboxes: list[BBox],
    pred_bboxes: list[BBox],
) -> list[list[float]] | npt.NDArray[np.float32]:
    """Calculates the [gt x pred] matrix of pairwise IoUs of GT and predicted bboxes of an image.

    Args:
        gt_bboxes (List[BBox]): GT bboxes
        pred_bboxes (List[BBox]): Predicted bboxes

    Returns:
        List[List[float]]: [gt x pred] matrix of pairwise IoUs of GT and predicted bboxes
    """

    ious = np.zeros((len(gt_bboxes), len(pred_bboxes)), dtype=float)

    for i, gt_bbox in enumerate(gt_bboxes):
        for j, pred_bbox in enumerate(pred_bboxes):
            ious[i, j] = calculate_bbox_iou(gt_bbox, pred_bbox)

    return ious


def calculate_bbox_iou(bbox_a: BBox, bbox_b: BBox) -> float:
    """Calculates the IoU between 2 bboxes.

    Args:
        bbox_a (BBox): BBox.
        bbox_b (BBox): BBox.

    Returns:
        float: IoU between the two bboxes.
    """

    intersect_area = bbox_a.intersection_area_with(bbox_b)
    union_area = bbox_a.union_area_with(bbox_b, intersect_area)

    return float(intersect_area) / float(union_area)


def calculate_iou_by_class(
    gt_bboxes: list[BBox],
    pred_bboxes: list[BBox],
    classes_count: int,
) -> list[float] | npt.NDArray[np.float32]:
    """Calculates the array of IoUs between GT and predicted bboxes of an image by class.
    This method considers bbox as segmentation masks to calculate the IoUs.

    Args:
        gt_bboxes (List[BBox]): GT bboxes
        pred_bboxes (List[BBox]): Predicted bboxes
        classes_count (int): Classes count

    Returns:
        List[float]: IoUs of an image indexed by class
    """

    ious = np.zeros(classes_count, dtype=float)

    for i in range(classes_count):
        ious[i] = __calculate_binary_iou(
            [gt_bbox for gt_bbox in gt_bboxes if gt_bbox.cls == i],
            [pred_bbox for pred_bbox in pred_bboxes if pred_bbox.cls == i],
        )

    return ious


def __calculate_binary_iou(
    gt_bboxes: list[BBox],
    pred_bboxes: list[BBox],
) -> float:
    if len(gt_bboxes) == 0 or len(pred_bboxes) == 0:
        return .0

    bboxes_btm_right_points = [bbox.bottom_right_point for bbox in gt_bboxes]\
        + [bbox.bottom_right_point for bbox in pred_bboxes]
    max_x = max(point[0] for point in bboxes_btm_right_points)
    max_y = max(point[1] for point in bboxes_btm_right_points)

    gt_mask = __draw_bboxes((int(max_x + 1), int(max_y + 1)), gt_bboxes)
    pred_mask = __draw_bboxes((int(max_x + 1), int(max_y + 1)), pred_bboxes)

    tp, f = (0, 0)
    flat_gt = __flat_mask(gt_mask)
    flat_pred = __flat_mask(pred_mask)
    for (curr_pred, curr_gt) in zip(flat_pred, flat_gt):
        if curr_gt == 1 and curr_pred == 1:
            tp += 1
        elif curr_gt != curr_pred:
            f += 1

    return tp / (f + tp)


def __draw_bboxes(mask_shape: tuple[int, int], bboxes: list[BBox]) -> list[list[int]] | npt.NDArray[np.uint]:
    mask = np.zeros(mask_shape, dtype=np.uint)

    for bbox in bboxes:
        mask[
            bbox.upper_left_point[0]:bbox.bottom_right_point[0] + 1,
            bbox.upper_left_point[1]:bbox.bottom_right_point[1] + 1,
        ] = 1

    return mask


def __flat_mask(mask: Mask | npt.NDArray[np.uint]) -> list[int]:
    return [item for sublist in mask for item in sublist]



from __future__ import annotations

from typing import Any

import pandas as pd
import shapely.wkt

from lapixdl.base import FileTypeError


class LapixDataFrame(pd.DataFrame):
    """An abstraction of pandas DataFrame for Lapix builtin type"""
    _metadata = ['geometry', 'category']

    @property
    def _constructor(self):
        return LapixDataFrame


def load(
    filename: str,
    **kwargs: Any,
) -> LapixDataFrame:
    """Loads the lapix parquet file into a Lapix dataframe

    Args:
        filename (str): The full filename for the parquet file.
        kwargs (Any): Kwargs for the read parquet method from pandas.

    Returns: A Lapix DataFrame from parquet file file.
    """
    if not filename.endswith(('.parquet.gzip', '.parquet')):
        raise FileTypeError('The file is not a parquet file.')

    df = pd.read_parquet(filename, **kwargs)

    lapix_df = LapixDataFrame(df)

    if 'geometry' in df.columns:
        # buffer(0) applied to fix invalid geometries. From shapely issue #278
        lapix_df['geometry'] = lapix_df['geometry'].apply(lambda x: shapely.wkt.loads(x).buffer(0))

    return lapix_df


def save(
        lapix_df: LapixDataFrame,
        filename: str,
        compression: str = 'gzip',
        **kwargs: Any,
) -> None:
    """Save a Lapix DataFrame into a parquet file

    Args:
        lapix_df (LapixDataFrame): The lapix dataframe to be saved.
        filename (str): The full filename for the parquet file.
        compression (str): The type of compression to be applied need to
    be: `snappy`, `gzip`, `brotli` or None
        kwargs (Any): Kwargs for the `to_parquet` method from pandas.
    """
    df_out = lapix_df.copy()
    df_out['geometry'] = df_out['geometry'].apply(lambda x: x.wkt)
    df_out.to_parquet(filename, compression=compression, **kwargs)


def generate_ids(
    column: pd.Series,
) -> pd.Series:
    column = column.astype('category')
    return column.cat.codes + 1


def geometries_area(
    lapix_df: LapixDataFrame,
) -> pd.Series:
    return pd.Series(lapix_df['geometry'].apply(lambda x: x.area))



from __future__ import annotations

import argparse
import sys
from typing import Sequence

from lapixdl.commands.converter import converter_command_parser


def main(argv: Sequence[str] | None = None) -> int:
    argv = argv if argv is not None else sys.argv[1:]

    parser = argparse.ArgumentParser(
        prog='LapixDL CLI tools',
        usage='lapixdl <command> [<args>]',
        description='This tool will help you to convert datasets between different formats.',
    )

    subparsers = parser.add_subparsers(
        dest='command',
        help='lapixdl command helpers',
    )

    # Register commands
    converter_command_parser(subparsers)

    help = subparsers.add_parser(
        'help',
        help='Show help for a specific command.',
    )
    help.add_argument(
        'help_cmd',
        nargs='?',
        help='Command to show help for.',
    )

    if len(argv) == 0:
        argv = ['help']

    args = parser.parse_args(argv)

    if args.command == 'help' and args.help_cmd:
        parser.parse_args([args.help_cmd, '--help'])
    elif args.command == 'help':
        parser.parse_args(['--help'])

    if not hasattr(args, 'func'):  # pragma: no cover
        parser.print_help()
        return 1

    return args.func(args)


if __name__ == '__main__':
    raise SystemExit(main())



from __future__ import annotations

import math
from typing import Union

import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import seaborn as sn
from matplotlib import cm
from matplotlib.axes import Axes
from matplotlib.colors import Colormap
from matplotlib.figure import Figure

from .model import Classification
from .model import Mask
from .model import Result
from lapixdl.formats.annotation import BBox

correct_color = sn.color_palette('Paired')[3]
incorrect_color = sn.color_palette('Paired')[5]

ColorMap = Union[str, Colormap]


def show_classifications(
        results: list[Result[Classification]],
        class_names: list[str],
        cols: int = 3,
        diff_correct_incorect: bool = True,
) -> tuple[Figure, Axes]:
    """Shows multiple classification results.

    Args:
        results (List[Result[Classification]]): List of classification results.
        class_names (List[str]): Class names.
        cols (int, optional): Number of colunms to show. Defaults to 3.
        diff_correct_incorect (bool, optional): Indicates if correct and incorrect
        results should be differentiated by color. Defaults to True.

    Returns:
        Tuple[Figure, Axes]: Figure and Axes of the ploted results.
    """

    rows = math.ceil(len(results) / cols)
    fig, axes = plt.subplots(rows, cols)
    if rows == 1:
        axes = [axes]
    fig.suptitle('Classifications')

    for i, result in enumerate(results):
        axe = axes[i // cols][i % cols]
        pred_s = (
            f'\nPred: {class_names[result.prediction.cls]} ({result.prediction.score})' if result.prediction is not None
            else ''
        )
        axe.set_title(
            f'GT: {class_names[result.gt.cls]}' + pred_s,
            fontsize='small',
            color='#333' if result.prediction is None or not diff_correct_incorect
            else (correct_color if result.prediction.cls == result.gt.cls else incorrect_color),
        )
        axe.axis('off')
        axe.imshow(result.image)

    for i in range(len(results), (cols * rows)):
        axe = axes[i // cols][i % cols]
        axe.axis('off')

    fig.set_size_inches((20, 10), forward=True)
    plt.tight_layout(w_pad=.2, h_pad=1.5)
    plt.show()
    return fig, axes


def show_segmentations(
        results: list[Result[Mask]],
        class_names: list[str],
        cmap: ColorMap = 'tab10',
        mask_alpha: float = .3,
) -> tuple[Figure, Axes]:
    """Shows segmentation results

    Args:
        results (List[Result[Mask]]): Segmentation results.
        class_names (List[str]): Class names.
        cmap (ColorMap, optional): Matplotlib color map for the mask or list of `class_index -> hex_color`. Defaults to 'tab10'.
        mask_alpha (float, optional): Alpha value for the mask colors. Defaults to .3.

    Returns:
        Tuple[Figure, Axes]: Figure and Axes of the ploted results.
    """

    cmap_colors = cm.get_cmap(cmap).colors

    assert len(cmap_colors) >= len(
        class_names,
    ), 'The color map length must be greater or equal the length of the class names.'

    rows = len(results)
    fig, axes = plt.subplots(rows, 3)
    if rows == 1:
        axes = [axes]

    fig.suptitle(' ', fontsize=40)  # To keep space for the legend
    lengend_handles = [
        mpatches.Patch(color=cmap_colors[i], label=code)
        for i, code in enumerate(class_names)
    ]
    fig.legend(
        handles=lengend_handles, fontsize='small',
        ncol=min(8, len(class_names)), loc='upper center',
    )

    for i, result in enumerate(results):
        axe_img = axes[i][0]
        axe_GT = axes[i][1]
        axe_pred = axes[i][2]

        axe_img.set_title('Image', fontsize='small')
        axe_img.imshow(result.image)
        axe_img.axis('off')

        axe_GT.set_title('GT', fontsize='small')
        axe_GT.axis('off')
        axe_GT.imshow(result.image)
        axe_GT.imshow(
            result.gt, cmap=cmap, alpha=mask_alpha,
            interpolation='none', vmin=0, vmax=len(cmap_colors) - 1,
        )

        axe_pred.axis('off')
        if result.prediction is not None:
            axe_pred.set_title('Prediction', fontsize='small')
            axe_pred.imshow(result.image)
            axe_pred.imshow(
                result.prediction, cmap=cmap, alpha=mask_alpha,
                interpolation='none', vmin=0, vmax=len(cmap_colors) - 1,
            )

    fig.set_size_inches((20, 10), forward=True)
    plt.tight_layout(w_pad=.2, h_pad=3)
    plt.show()

    return fig, axes


def show_detections(
    results: list[Result[list[BBox]]],
    class_names: list[str],
    cmap: ColorMap = 'tab10',
    show_bbox_label: bool = True,
) -> tuple[Figure, Axes]:
    """Shows detection results.

    Args:
        results (List[Result[List[BBox]]]): Detection results.
        class_names (List[str]): Class names.
        cmap (ColorMap, optional): Matplotlib color map for the mask or list of `class_index -> hex_color`. Defaults to 'tab10'.
        show_bbox_label (bool, optional): Indicates if the class label should be shown for each bbox. Defaults to True.

    Returns:
        Tuple[Figure, Axes]: Figure and Axes of the ploted results.
    """

    cmap_colors = cm.get_cmap(cmap).colors

    assert len(cmap_colors) >= len(
        class_names,
    ), 'The color map length must be greater or equal the length of the class names.'

    rows = len(results)
    fig, axes = plt.subplots(rows, 3)
    if rows == 1:
        axes = [axes]

    fig.suptitle(' ', fontsize=40)  # To keep space for the legend
    lengend_handles = [
        mpatches.Patch(color=cmap_colors[i], label=code)
        for i, code in enumerate(class_names)
    ]
    fig.legend(
        handles=lengend_handles, fontsize='small',
        ncol=min(8, len(class_names)), loc='upper center',
    )

    for i, result in enumerate(results):
        axe_img = axes[i][0]
        axe_GT = axes[i][1]
        axe_pred = axes[i][2]

        axe_img.set_title('Image', fontsize='small')
        axe_img.imshow(result.image)
        axe_img.axis('off')

        axe_GT.set_title('GT', fontsize='small')
        axe_GT.axis('off')
        axe_GT.imshow(result.image)
        __draw_bboxes(
            axe_GT, result.gt, cmap_colors,
            class_names, show_bbox_label,
        )

        axe_pred.axis('off')
        if result.prediction is not None:
            axe_pred.set_title('Prediction', fontsize='small')
            axe_pred.imshow(result.image)
            __draw_bboxes(
                axe_pred, result.prediction,
                cmap_colors, class_names, show_bbox_label,
            )

    fig.set_size_inches((20, 10), forward=True)
    plt.tight_layout(w_pad=.2, h_pad=3)
    plt.show()

    return fig, axes


def __draw_bboxes(axe: plt.Axes, bboxes: list[BBox], cmap_colors: list[str], class_names: list[str], show_bbox_label: bool):
    for bbox in bboxes:
        color = cmap_colors[bbox.cls]
        label = (f'{class_names[bbox.cls]}' if show_bbox_label else '') + \
            (f' ({round(bbox.score, 3)})' if bbox.score else '')
        rect = mpatches.Rectangle(
            bbox.upper_left_point,
            bbox.width, bbox.height,
            linewidth=2,
            edgecolor=color,
            facecolor='none',
        )
        axe.add_patch(rect)
        axe.annotate(
            label.strip(),
            (
                bbox.upper_left_point[0] + 5,
                bbox.upper_left_point[1] - 5,
            ),
            color='w',
            weight='bold',
            fontsize=10,
            ha='left', va='bottom',
            bbox=dict(facecolor=color, edgecolor='none', pad=1.5),
        )



from __future__ import annotations

from copy import copy
from dataclasses import dataclass

from shapely.geometry import MultiPolygon
from shapely.geometry import Polygon


@dataclass
class BBox:
    """Bounding Box data structure

    Attributes:
        upper_left_x (int): Upper left X position of the Bounding Box.
        upper_left_y (int): Upper left Y position of the Bounding Box.
        width (int): Width of the Bounding Box.
        height (int): Height of the Bounding Box.
        cls (int): Bounding Box class index.
        score (Optional[float]): Bounding Box prediction score.
    """
    upper_left_x: int
    upper_left_y: int
    width: int
    height: int
    cls: int
    score: float | None = None

    def __post_init__(self):
        if self.upper_left_x < 0 or self.upper_left_y < 0:
            raise ValueError(f'The upper left (x, y) should be positive values. Got ({self.upper_left_x}, {self.upper_left_y})')

        if self.width <= 0:
            raise ValueError(f'The width should be bigger than zero. Got {self.width}')

        if self.height <= 0:
            raise ValueError(f'The height should be bigger than zero. Got {self.height}')

    @property
    def upper_left_point(self) -> tuple[int, int]:
        """Tuple[int, int]: (X,Y) of the upper left point of the Bounding Box."""
        return (
            self.upper_left_x,
            self.upper_left_y,
        )

    @property
    def upper_right_point(self) -> tuple[int, int]:
        """Tuple[int, int]: (X,Y) of the upper right point of the Bounding Box."""
        return (
            self.upper_left_x + self.width - 1,
            self.upper_left_y,
        )

    @property
    def bottom_right_point(self) -> tuple[int, int]:
        """Tuple[int, int]: (X,Y) of the bottom right point of the Bounding Box."""
        return (
            self.upper_left_x + self.width - 1,
            self.upper_left_y + self.height - 1,
        )

    @property
    def bottom_left_point(self) -> tuple[int, int]:
        """Tuple[int, int]: (X,Y) of the bottom left point of the Bounding Box."""
        return (
            self.upper_left_x,
            self.upper_left_y + self.height - 1,
        )

    @property
    def center_point(self) -> tuple[int, int]:
        """Tuple[int, int]: (X,Y) of the center point of the Bounding Box."""
        return (
            self.upper_left_x + ((self.width - 1) // 2),
            self.upper_left_y + ((self.height - 1) // 2),
        )

    @property
    def area(self) -> int:
        """int: Area of the Bounding Box."""
        return self.width * self.height

    @property
    def coords(self) -> tuple[tuple[int, int], ...]:
        """Tuple[Tuple[int, int], ...]: A tuple with each point of the Bounding Box."""
        return (
            self.upper_left_point,
            self.upper_right_point,
            self.bottom_right_point,
            self.bottom_left_point,
        )

    @property
    def xy(self) -> tuple[list[int], list[int]]:
        """Tuple[list[int], list[int]]: A tuple with lists of coords for each x and y"""
        _x, _y = zip(*self.coords)
        return (list(_x), list(_y))

    @property
    def slice_x(self) -> slice:
        """slice: A slice between upper_left_x and upper_right_x"""
        return slice(self.upper_left_x, self.upper_right_point[0])

    @property
    def slice_y(self) -> slice:
        """slice: A slice between upper_left_y and bottom_left_y"""
        return slice(self.upper_left_y, self.bottom_left_point[1])

    def intersection_area_with(self: BBox, bbox: BBox) -> int:
        """Calculates the intersection area with another bbox

        Args:
            self (BBox): This bbox
            bbox (BBox): Bbox to intersect with

        Returns:
            int: The intersection area with the bbox
        """

        # Gets each box upper left and bottom right coordinates
        (upr_lft_x_a, upr_lft_y_a) = self.upper_left_point
        (btm_rgt_x_a, btm_rgt_y_a) = self.bottom_right_point

        (upr_lft_x_b, upr_lft_y_b) = bbox.upper_left_point
        (btm_rgt_x_b, btm_rgt_y_b) = bbox.bottom_right_point

        # Calculates the intersection box upper left and bottom right coordinates
        (upr_lft_x_intersect, upr_lft_y_intersect) = (
            max(upr_lft_x_a, upr_lft_x_b), max(upr_lft_y_a, upr_lft_y_b),
        )
        (btm_rgt_x_intersect, btm_rgt_y_intersect) = (
            min(btm_rgt_x_a, btm_rgt_x_b), min(btm_rgt_y_a, btm_rgt_y_b),
        )

        # Calculates the height and width of the intersection box
        (w_intersect, h_intersect) = (
            btm_rgt_x_intersect - upr_lft_x_intersect + 1,
            btm_rgt_y_intersect - upr_lft_y_intersect + 1,
        )

        # If H or W <= 0, there is no intersection
        if (w_intersect <= 0) or (h_intersect <= 0):
            return 0

        return w_intersect * h_intersect

    def union_area_with(
            self: BBox,
            bbox: BBox,
            intersection_area: int | None = None,
    ) -> int:
        """Calculates the union area with another bbox

        Args:
            self (BBox): This bbox
            bbox (BBox): Bbox to union with
            intersection_area (Optional[int], optional): The intersection area between this and bbox. Defaults to None.

        Returns:
            int: The union area with the bbox
        """

        return self.area + bbox.area - (intersection_area or self.intersection_area_with(bbox))

    def to_polygon(self) -> Polygon:
        """Polygon: A shapely polygon from the coordinates of the bbox"""
        return Polygon(self.coords)


@dataclass
class Annotation:
    geometry: Polygon | MultiPolygon
    category_id: int
    iscrowd: int = 0

    @property
    def bbox(self) -> BBox:
        return bounds_to_bbox(self.geometry.bounds, self.category_id)

    @property
    def _geo_type(self) -> str:
        return self.geometry.geom_type

    @property
    def xywh_bbox(self) -> list[int]:
        bbox = self.bbox
        return [bbox.upper_left_x, bbox.upper_left_y, bbox.width, bbox.height]

    def __iter__(self) -> Annotation:
        self._idx = 0

        if self._geo_type == 'MultiPolygon':
            self._geometries = list(self.geometry.geoms)
        elif self._geo_type == 'Polygon':
            self._geometries = [self.geometry]
        else:
            raise TypeError(f'Unexpected geometry type (`{self._geo_type}`) - expected `MultiPolygon` or `Polygon`')

        return self

    def __next__(self) -> Polygon:
        if self._idx < len(self._geometries):
            out = self._geometries[self._idx]
            self._idx += 1
            return out
        else:
            raise StopIteration

    def copy(self) -> Annotation:
        return copy(self)


def bounds_to_bbox(bounds: tuple[float], category_id: int) -> BBox:
    """Generate a BBox from a tuple of bounds

    Args:
        bounds (tuple[float]): A tuple of floats in the following order
    (min_x, min_y, max_x, max_y)
        category_id: A category id fot the BBox

    Returns:
        BBox: An BBox based on the bounds
    """
    b = tuple(int(i) for i in bounds)
    min_x, min_y, max_x, max_y = b
    return BBox(
        upper_left_x=min_x,
        upper_left_y=min_y,
        width=max_x - min_x,
        height=max_y - min_y,
        cls=category_id,
    )



from __future__ import annotations

import json
from typing import Any

import pandas as pd


EXPECTED_DATA = set({'ID', 'External ID', 'Reviews', 'Label'})


def load(
    filename: str,
) -> pd.DataFrame:
    """Loads the labelbox JSON file into a pandas dataframe

    Args
        filename (str): The full filename for the JSON file.

    Returns: A pandas DataFrame with the data from labelbox JSON file.
    """
    with open(filename) as f:
        raw = json.load(f)

    validate(raw)

    return pd.DataFrame(raw)


def validate(
    raw: list[dict[str, Any]],
) -> bool:
    """Validate if the content of labelbox file is valid

    Args:
        raw (list[dict[str, Any]]): A list with the content from the
    labelbox JSON file.

    Returns:
        bool: True if the content data is valid.
    """
    if not isinstance(raw, list):
        raise TypeError('A list of dictionaries representing the raw data from LabelBox was expected.')

    for it in raw:
        if not all(i in it for i in EXPECTED_DATA) and 'Skipped' not in it:
            raise KeyError(f'The annotation values need to have `Skipped` or all of {EXPECTED_DATA} keys.')

    return True
